#include "core/VulkanWrapper.hpp"

#include <core/File.hpp>
#include <core/UniquePointer.hpp>

#include "GLFW.hpp"
#include "core/VulkanBase.hpp"

namespace Vulkan = Core::Vulkan;
using Vulkan::CommandBuffer;
using Vulkan::SwapchainImages;

#define WRAPPER_DESTRUCT(Type, ...)                                \
    using Core::Vulkan::Type;                                      \
    Type::~Type() {                                                \
        if(device != VK_NULL_HANDLE) {                             \
            vkDestroy##Type##__VA_ARGS__(device, handle, nullptr); \
        }                                                          \
    }

WRAPPER_DESTRUCT(ImageView)
WRAPPER_DESTRUCT(Framebuffer)
WRAPPER_DESTRUCT(Semaphore)
WRAPPER_DESTRUCT(Fence)
WRAPPER_DESTRUCT(Swapchain, KHR)
WRAPPER_DESTRUCT(ShaderModule)
WRAPPER_DESTRUCT(PipelineLayout)

bool ImageView::init(VkDevice d, VkImage image, VkFormat format) {
    VkImageViewCreateInfo info = {
        .sType = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO,
        .image = image,
        .viewType = VK_IMAGE_VIEW_TYPE_2D,
        .format = format,
        .subresourceRange = {
            .aspectMask = VK_IMAGE_ASPECT_COLOR_BIT,
            .baseMipLevel = 0,
            .levelCount = 1,
            .baseArrayLayer = 0,
            .layerCount = 1}};
    VK_CHECK_TRUE(vkCreateImageView(d, &info, nullptr, &handle));
    device = d;
    return false;
}

bool Framebuffer::init(
    const ImageView& iv, VkRenderPass rp, u32 width, u32 height) {
    VkFramebufferCreateInfo info = {
        .sType = VK_STRUCTURE_TYPE_FRAMEBUFFER_CREATE_INFO,
        .renderPass = rp,
        .attachmentCount = 1,
        .pAttachments = iv,
        .width = width,
        .height = height,
        .layers = 1};
    VK_CHECK_TRUE(vkCreateFramebuffer(iv.device, &info, nullptr, &handle));
    device = iv.device;
    return false;
}

bool Semaphore::init(VkDevice d) {
    VkSemaphoreCreateInfo info = {
        .sType = VK_STRUCTURE_TYPE_SEMAPHORE_CREATE_INFO};
    VK_CHECK_TRUE(vkCreateSemaphore(d, &info, nullptr, &handle));
    device = d;
    return false;
}

bool Fence::init(VkDevice d) {
    VkFenceCreateInfo info = {
        .sType = VK_STRUCTURE_TYPE_FENCE_CREATE_INFO,
        .flags = VK_FENCE_CREATE_SIGNALED_BIT};
    VK_CHECK_TRUE(vkCreateFence(d, &info, nullptr, &handle));
    device = d;
    return false;
}

static u32 getSwapImageCount(const VkSurfaceCapabilitiesKHR* caps) {
    u32 c = caps->minImageCount + 1;
    // according to VkSurfaceCapabilitiesKHR doc:
    // maxImageCount is 0 when there is no strict limit
    if(caps->maxImageCount != 0) {
        return Core::min(c, caps->maxImageCount);
    }
    return c;
}

bool Swapchain::init(Data& d) {
    VkSurfaceCapabilitiesKHR caps = {0};
    if(d.base.getSurfaceCapabilities(caps)) {
        return true;
    }
    VkSwapchainCreateInfoKHR ci = {
        .sType = VK_STRUCTURE_TYPE_SWAPCHAIN_CREATE_INFO_KHR,
        .surface = d.base,
        .minImageCount = getSwapImageCount(&caps),
        .imageFormat = d.surfaceFormat.format,
        .imageColorSpace = d.surfaceFormat.colorSpace,
        .imageExtent = d.size,
        .imageArrayLayers = 1,
        .imageUsage = VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT,
        .imageSharingMode = d.sharingMode,
        .queueFamilyIndexCount = static_cast<u32>(d.queueFamilies.getLength()),
        .pQueueFamilyIndices = &d.queueFamilies[0],
        .preTransform = caps.currentTransform,
        .compositeAlpha = VK_COMPOSITE_ALPHA_OPAQUE_BIT_KHR,
        .presentMode = d.presentMode,
        .clipped = VK_TRUE,
        .oldSwapchain = VK_NULL_HANDLE};
    VK_CHECK_TRUE(vkCreateSwapchainKHR(d.base, &ci, nullptr, &handle));
    device = d.base;
    return false;
}

bool ShaderModule::init(VkDevice d, const char* path) {
    List<char> f;
    if(Core::readFile(f, path)) {
        return true;
    }
    size_t l = f.getLength() - 1;
    if((l % 4) != 0) {
        VK_REPORT_ERROR("Shader size is not a multiple of 4");
        return true;
    }
    VkShaderModuleCreateInfo info = {
        .sType = VK_STRUCTURE_TYPE_SHADER_MODULE_CREATE_INFO,
        .codeSize = l,
        .pCode = reinterpret_cast<u32*>(static_cast<void*>(&f[0]))};
    VK_CHECK_TRUE(vkCreateShaderModule(d, &info, nullptr, &handle));
    device = d;
    return false;
}

bool PipelineLayout::init(VkDevice d) {
    VkPipelineLayoutCreateInfo info = {
        .sType = VK_STRUCTURE_TYPE_PIPELINE_LAYOUT_CREATE_INFO};
    VK_CHECK_TRUE(vkCreatePipelineLayout(d, &info, nullptr, &handle));
    device = d;
    return false;
}

bool SwapchainImages::init(const Swapchain& s, VkFormat format) {
    u32 c = 0;
    VK_CHECK_TRUE(vkGetSwapchainImagesKHR(s.device, s.handle, &c, nullptr));
    images.resize(c);
    imageViews.resize(c);
    VK_CHECK_TRUE(vkGetSwapchainImagesKHR(s.device, s.handle, &c, &images[0]));
    for(u32 x = 0; x < c; x++) {
        if(imageViews[x].init(s.device, images[x], format)) {
            return true;
        }
    }
    return false;
}

bool CommandBuffer::init(VkDevice d, VkCommandPool cp) {
    VkCommandBufferAllocateInfo info = {
        .sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO,
        .commandPool = cp,
        .level = VK_COMMAND_BUFFER_LEVEL_PRIMARY,
        .commandBufferCount = 1};
    VK_CHECK_TRUE(vkAllocateCommandBuffers(d, &info, &handle));
    device = d;
    return false;
}

struct VulkanDummy {
    Core::Vulkan::Base base{};

    struct BaseData {
        u32 graphicsFamily = 0;
        u32 presentFamily = 0;
        VkSurfaceFormatKHR surfaceFormat{};
        VkPresentModeKHR presentMode{};
    };

    BaseData baseData{};
    VkQueue graphicsQueue{};
    VkQueue presentQueue{};
    VkExtent2D swapchainSize{};
    Swapchain swapchain{};
    SwapchainImages images{};
    ShaderModule vertexShaderModule{};
    ShaderModule fragmentShaderModule{};
    PipelineLayout pipelineLayout{};
    VkRenderPass renderPass{};
    VkViewport viewport{};
    VkRect2D scissor{};
    VkPipeline pipeline{};
    Core::List<Vulkan::Framebuffer> framebuffers{};
    VkCommandPool commandPool{};
    static constexpr size_t MAX_FRAMES = 2;
    size_t currentFrame = 0;
    bool shouldWait = false;

    struct Frame {
        CommandBuffer commandBuffer{};
        Semaphore imageAvailableSemaphore{};
        Semaphore renderFinishedSemaphore{};
        Fence inFlightFence{};
    };

    Frame frames[MAX_FRAMES];

    static int getSurfaceFormatPoints(const VkSurfaceFormatKHR& sf) {
        if(sf.format == VK_FORMAT_B8G8R8A8_UNORM &&
           sf.colorSpace == VK_COLOR_SPACE_SRGB_NONLINEAR_KHR) {
            return 10;
        }
        return 1;
    }

    static int getSurfacePresentModePoints(VkPresentModeKHR m) {
        if(m == VK_PRESENT_MODE_MAILBOX_KHR) {
            return 5;
        } else if(m == VK_PRESENT_MODE_FIFO_KHR) {
            return 2;
        }
        return 0;
    }

    bool getSwapchainSize(VkExtent2D* size) {
        VkSurfaceCapabilitiesKHR c = {0};
        if(base.getSurfaceCapabilities(c)) {
            return true;
        }
        if(c.currentExtent.width != 0xFFFF'FFFFu &&
           c.currentExtent.height != 0xFFFF'FFFFu) {
            *size = c.currentExtent;
            LOG_INFO("Swapchain size: #x#", size->width, size->height);
            return false;
        }
        int w = 0;
        int h = 0;
        glfwGetFramebufferSize(Core::Window::get(), &w, &h);
        if(w <= 0 || h <= 0) {
            LOG_ERROR("Could not get framebuffer size");
            return true;
        }
        LOG_INFO("Framebuffer size: #x#", w, h);
        size->width = Core::clamp(
            static_cast<u32>(w), c.minImageExtent.width,
            c.maxImageExtent.width);
        size->height = Core::clamp(
            static_cast<u32>(h), c.minImageExtent.height,
            c.maxImageExtent.height);
        LOG_INFO("Swapchain size: #x#", size->width, size->height);
        return false;
    }

    bool initSwapchain() {
        Vulkan::Swapchain::Data d = {
            .base = base,
            .surfaceFormat = baseData.surfaceFormat,
            .presentMode = baseData.presentMode};
        if(getSwapchainSize(&d.size)) {
            LOG_ERROR("Could not retrieve any swapchain size");
            return true;
        }
        swapchainSize = d.size;
        d.queueFamilies.add(baseData.graphicsFamily);
        if(baseData.graphicsFamily != baseData.presentFamily) {
            d.queueFamilies.add(baseData.presentFamily);
            d.sharingMode = VK_SHARING_MODE_CONCURRENT;
        } else {
            d.sharingMode = VK_SHARING_MODE_EXCLUSIVE;
        }
        return swapchain.init(d);
    }

    bool initDevice() {
        bool same = baseData.graphicsFamily == baseData.presentFamily;
        Core::List<Vulkan::DeviceQueueData> data;
        data.add(baseData.graphicsFamily, 1.0f);
        if(!same) {
            data.add(baseData.presentFamily, 1.0f);
        };
        Core::List<const char*> extensions;
        extensions.add(VK_KHR_SWAPCHAIN_EXTENSION_NAME);
        if(base.initDevice(data, extensions)) {
            return true;
        }
        vkGetDeviceQueue(base, baseData.graphicsFamily, 0, &graphicsQueue);
        if(graphicsQueue == VK_NULL_HANDLE) {
            LOG_ERROR("Cannot get device graphics queue");
            return true;
        }
        if(same) {
            presentQueue = graphicsQueue;
            return false;
        }
        vkGetDeviceQueue(base, baseData.presentFamily, 0, &presentQueue);
        if(presentQueue == VK_NULL_HANDLE) {
            LOG_ERROR("Cannot get device present queue");
            return true;
        }
        return false;
    }

    static int getDevicePoints(Vulkan::Base& b, BaseData& d) {
        int points = 0;
        VkPhysicalDeviceProperties p;
        b.getPhysicalDeviceProperties(p);
        LOG_INFO("Checking '#'", p.deviceName);
        switch(p.deviceType) {
            case VK_PHYSICAL_DEVICE_TYPE_DISCRETE_GPU: points += 100; break;
            case VK_PHYSICAL_DEVICE_TYPE_INTEGRATED_GPU: points += 50; break;
            case VK_PHYSICAL_DEVICE_TYPE_VIRTUAL_GPU: points += 20; break;
            default: break;
        }
        d.graphicsFamily = b.findQueueFamily(VK_QUEUE_GRAPHICS_BIT);
        if(d.graphicsFamily == Vulkan::INVALID_QUEUE_FAMILY) {
            LOG_INFO("> ... has no graphics family");
            points = -1;
        }
        d.presentFamily = b.findSurfaceQueueFamily();
        if(d.presentFamily == Vulkan::INVALID_QUEUE_FAMILY) {
            LOG_INFO("> ... has no present family");
            points = -1;
        }
        if(!b.hasExtension(VK_KHR_SWAPCHAIN_EXTENSION_NAME)) {
            LOG_INFO("> ... has no swapchain support");
            points = -1;
        }
        if(b.findSurfaceFormat(d.surfaceFormat, getSurfaceFormatPoints)) {
            LOG_INFO("> ... has no matching surface format");
            points = -1;
        } else {
            points += getSurfaceFormatPoints(d.surfaceFormat);
        }
        if(b.findSurfacePresentMode(
               d.presentMode, getSurfacePresentModePoints)) {
            LOG_INFO("> ... has no matching present mode");
            points = -1;
        } else {
            points += getSurfacePresentModePoints(d.presentMode);
        }
        LOG_INFO("> Final points: #", points);
        return points;
    }

    bool initPhysicalDevice() {
        LOG_INFO("Searching for physical devices ...");
        if(base.findPhysicalDevice(baseData, getDevicePoints)) {
            LOG_ERROR("No matching physical device was found");
            return true;
        }
        VkPhysicalDeviceProperties p;
        base.getPhysicalDeviceProperties(p);
        LOG_INFO("Best Device: #", p.deviceName);
        return false;
    }

    bool initSwapchainImages() {
        if(images.init(swapchain, baseData.surfaceFormat.format)) {
            LOG_ERROR("Could not get swapchain images");
            return true;
        }
        LOG_INFO("Found # images", images.images.getLength());
        return false;
    }

    bool initShaders() {
        return vertexShaderModule.init(base, "shaders/vertex.spv") ||
               fragmentShaderModule.init(base, "shaders/fragment.spv");
    }

    bool initPipeline() {
        VkPipelineShaderStageCreateInfo stages[2] = {
            {.sType = VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO,
             .stage = VK_SHADER_STAGE_VERTEX_BIT,
             .module = vertexShaderModule,
             .pName = "main"},
            {.sType = VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO,
             .stage = VK_SHADER_STAGE_FRAGMENT_BIT,
             .module = fragmentShaderModule,
             .pName = "main"}};

        VkPipelineVertexInputStateCreateInfo vertexInputState = {
            .sType = VK_STRUCTURE_TYPE_PIPELINE_VERTEX_INPUT_STATE_CREATE_INFO};

        VkPipelineInputAssemblyStateCreateInfo inputAssemblyState = {
            .sType =
                VK_STRUCTURE_TYPE_PIPELINE_INPUT_ASSEMBLY_STATE_CREATE_INFO,
            .topology = VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST,
            .primitiveRestartEnable = false};

        viewport = {
            .x = 0.0f,
            .y = 0.0f,
            .width = static_cast<float>(swapchainSize.width),
            .height = static_cast<float>(swapchainSize.height),
            .minDepth = 0.0f,
            .maxDepth = 1.0f};
        scissor = {.offset = {0, 0}, .extent = swapchainSize};
        VkPipelineViewportStateCreateInfo viewportState = {
            .sType = VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_STATE_CREATE_INFO,
            .viewportCount = 1,
            .pViewports = &viewport,
            .scissorCount = 1,
            .pScissors = &scissor};

        VkPipelineRasterizationStateCreateInfo rasterizationState = {
            .sType = VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_STATE_CREATE_INFO,
            .depthClampEnable = false,
            .rasterizerDiscardEnable = false,
            .polygonMode = VK_POLYGON_MODE_FILL,
            .cullMode = VK_CULL_MODE_BACK_BIT,
            .frontFace = VK_FRONT_FACE_CLOCKWISE,
            .depthBiasEnable = false,
            .depthBiasConstantFactor = 0.0f,
            .depthBiasClamp = 0.0f,
            .depthBiasSlopeFactor = 0.0f,
            .lineWidth = 1.0f};

        VkPipelineMultisampleStateCreateInfo multisampleState = {
            .sType = VK_STRUCTURE_TYPE_PIPELINE_MULTISAMPLE_STATE_CREATE_INFO,
            .rasterizationSamples = VK_SAMPLE_COUNT_1_BIT,
            .sampleShadingEnable = false,
            .minSampleShading = 1.0f,
            .pSampleMask = nullptr,
            .alphaToCoverageEnable = false,
            .alphaToOneEnable = false};

        VkPipelineColorBlendAttachmentState colorBlendAttachmentState = {
            .blendEnable = false,
            .srcColorBlendFactor = VK_BLEND_FACTOR_ONE,
            .dstColorBlendFactor = VK_BLEND_FACTOR_ZERO,
            .colorBlendOp = VK_BLEND_OP_ADD,
            .srcAlphaBlendFactor = VK_BLEND_FACTOR_ONE,
            .dstAlphaBlendFactor = VK_BLEND_FACTOR_ZERO,
            .alphaBlendOp = VK_BLEND_OP_ADD,
            .colorWriteMask =
                VK_COLOR_COMPONENT_R_BIT | VK_COLOR_COMPONENT_G_BIT |
                VK_COLOR_COMPONENT_B_BIT | VK_COLOR_COMPONENT_A_BIT};

        VkPipelineColorBlendStateCreateInfo colorBlendState = {
            .sType = VK_STRUCTURE_TYPE_PIPELINE_COLOR_BLEND_STATE_CREATE_INFO,
            .logicOpEnable = false,
            .logicOp = VK_LOGIC_OP_COPY,
            .attachmentCount = 1,
            .pAttachments = &colorBlendAttachmentState,
            .blendConstants = {0.0f, 0.0f, 0.0f, 0.0f}};

        Core::Array<VkDynamicState, 2> dynamicStates = {
            {VK_DYNAMIC_STATE_VIEWPORT, VK_DYNAMIC_STATE_SCISSOR}};

        VkPipelineDynamicStateCreateInfo dynamicState = {
            .sType = VK_STRUCTURE_TYPE_PIPELINE_DYNAMIC_STATE_CREATE_INFO,
            .dynamicStateCount = dynamicStates.getLength(),
            .pDynamicStates = dynamicStates.begin()};

        VkGraphicsPipelineCreateInfo info = {
            .sType = VK_STRUCTURE_TYPE_GRAPHICS_PIPELINE_CREATE_INFO,
            .stageCount = 2,
            .pStages = stages,
            .pVertexInputState = &vertexInputState,
            .pInputAssemblyState = &inputAssemblyState,
            .pTessellationState = nullptr,
            .pViewportState = &viewportState,
            .pRasterizationState = &rasterizationState,
            .pMultisampleState = &multisampleState,
            .pDepthStencilState = nullptr,
            .pColorBlendState = &colorBlendState,
            .pDynamicState = &dynamicState,
            .layout = pipelineLayout,
            .renderPass = renderPass,
            .subpass = 0,
            .basePipelineHandle = VK_NULL_HANDLE,
            .basePipelineIndex = -1};

        VK_CHECK_TRUE(vkCreateGraphicsPipelines(
            base, VK_NULL_HANDLE, 1, &info, nullptr, &pipeline));
        return false;
    }

    bool initFramebuffers() {
        framebuffers.resize(images.imageViews.getLength());
        size_t i = 0;
        for(Framebuffer& f : framebuffers) {
            if(f.init(
                   images.imageViews[i++], renderPass, swapchainSize.width,
                   swapchainSize.height)) {
                return true;
            }
        }
        return false;
    }

    bool initRenderPass() {
        VkAttachmentDescription c = {
            .format = baseData.surfaceFormat.format,
            .samples = VK_SAMPLE_COUNT_1_BIT,
            .loadOp = VK_ATTACHMENT_LOAD_OP_CLEAR,
            .storeOp = VK_ATTACHMENT_STORE_OP_STORE,
            .stencilLoadOp = VK_ATTACHMENT_LOAD_OP_DONT_CARE,
            .stencilStoreOp = VK_ATTACHMENT_STORE_OP_DONT_CARE,
            .initialLayout = VK_IMAGE_LAYOUT_UNDEFINED,
            .finalLayout = VK_IMAGE_LAYOUT_PRESENT_SRC_KHR};
        VkAttachmentReference ca = {
            .attachment = 0,
            .layout = VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL};
        VkSubpassDescription subpass = {
            .pipelineBindPoint = VK_PIPELINE_BIND_POINT_GRAPHICS,
            .colorAttachmentCount = 1,
            .pColorAttachments = &ca};
        VkSubpassDependency dependency = {
            .srcSubpass = VK_SUBPASS_EXTERNAL,
            .dstSubpass = 0,
            .srcStageMask = VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT,
            .dstStageMask = VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT,
            .srcAccessMask = 0,
            .dstAccessMask = VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT,
        };
        VkRenderPassCreateInfo info = {
            .sType = VK_STRUCTURE_TYPE_RENDER_PASS_CREATE_INFO,
            .attachmentCount = 1,
            .pAttachments = &c,
            .subpassCount = 1,
            .pSubpasses = &subpass,
            .dependencyCount = 1,
            .pDependencies = &dependency};
        VK_CHECK_TRUE(vkCreateRenderPass(base, &info, nullptr, &renderPass));
        return false;
    }

    bool initCommandPool() {
        VkCommandPoolCreateInfo info = {
            .sType = VK_STRUCTURE_TYPE_COMMAND_POOL_CREATE_INFO,
            .flags = VK_COMMAND_POOL_CREATE_RESET_COMMAND_BUFFER_BIT,
            .queueFamilyIndex = baseData.graphicsFamily};
        VK_CHECK_TRUE(vkCreateCommandPool(base, &info, nullptr, &commandPool));
        return false;
    }

    bool fillCommandBuffer(VkCommandBuffer cb, u32 index) {
        VkCommandBufferBeginInfo info = {
            .sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO};
        VK_CHECK_TRUE(vkBeginCommandBuffer(cb, &info));
        VkClearValue v = {.color = {.float32 = {0.0f, 0.5f, 0.0f, 1.0f}}};
        VkRenderPassBeginInfo rInfo = {
            .sType = VK_STRUCTURE_TYPE_RENDER_PASS_BEGIN_INFO,
            .renderPass = renderPass,
            .framebuffer = framebuffers[index],
            .renderArea = {.offset = {0, 0}, .extent = swapchainSize},
            .clearValueCount = 1,
            .pClearValues = &v};
        vkCmdBeginRenderPass(cb, &rInfo, VK_SUBPASS_CONTENTS_INLINE);
        vkCmdBindPipeline(cb, VK_PIPELINE_BIND_POINT_GRAPHICS, pipeline);
        vkCmdSetViewport(cb, 0, 1, &viewport);
        vkCmdSetScissor(cb, 0, 1, &scissor);
        vkCmdDraw(cb, 3, 1, 0, 0);
        vkCmdEndRenderPass(cb);
        VK_CHECK_TRUE(vkEndCommandBuffer(cb));
        return false;
    }

    bool initFrames() {
        for(size_t i = 0; i < MAX_FRAMES; i++) {
            if(frames[i].commandBuffer.init(base, commandPool) ||
               frames[i].imageAvailableSemaphore.init(base) ||
               frames[i].renderFinishedSemaphore.init(base) ||
               frames[i].inFlightFence.init(base)) {
                return true;
            }
        }
        return false;
    }

    bool init() {
        return base.init() || initPhysicalDevice() || initDevice() ||
               initSwapchain() || initSwapchainImages() || initShaders() ||
               pipelineLayout.init(base) || initRenderPass() ||
               initPipeline() || initFramebuffers() || initCommandPool() ||
               initFrames();
    }

    bool render() {
        if(shouldWait) {
            return false;
        }
        Frame* f = frames + currentFrame;
        VK_CHECK_TRUE(
            vkWaitForFences(base, 1, f->inFlightFence, true, UINT64_MAX));
        VK_CHECK_TRUE(vkResetFences(base, 1, f->inFlightFence));

        uint32_t imageIndex = 0;
        VK_CHECK_TRUE(vkAcquireNextImageKHR(
            base, swapchain, UINT64_MAX, f->imageAvailableSemaphore,
            VK_NULL_HANDLE, &imageIndex));

        vkResetCommandBuffer(f->commandBuffer, 0);
        fillCommandBuffer(f->commandBuffer, imageIndex);

        VkSemaphore waitSemaphores[] = {f->imageAvailableSemaphore};
        VkPipelineStageFlags waitStages[] = {
            VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT};
        VkSemaphore signalSemaphores[] = {f->renderFinishedSemaphore};
        VkSubmitInfo info = {
            .sType = VK_STRUCTURE_TYPE_SUBMIT_INFO,
            .waitSemaphoreCount = 1,
            .pWaitSemaphores = waitSemaphores,
            .pWaitDstStageMask = waitStages,
            .commandBufferCount = 1,
            .pCommandBuffers = f->commandBuffer,
            .signalSemaphoreCount = 1,
            .pSignalSemaphores = signalSemaphores};
        VK_CHECK_TRUE(vkQueueSubmit(graphicsQueue, 1, &info, f->inFlightFence));

        VkPresentInfoKHR presentInfo = {
            .sType = VK_STRUCTURE_TYPE_PRESENT_INFO_KHR,
            .waitSemaphoreCount = 1,
            .pWaitSemaphores = signalSemaphores,
            .swapchainCount = 1,
            .pSwapchains = swapchain,
            .pImageIndices = &imageIndex};

        VK_CHECK_TRUE(vkQueuePresentKHR(presentQueue, &presentInfo));
        currentFrame = (currentFrame + 1) % MAX_FRAMES;
        return false;
    }

    void destroy() {
        if(static_cast<VkDevice>(base) != VK_NULL_HANDLE) {
            vkDeviceWaitIdle(base);
            vkDestroyCommandPool(base, commandPool, nullptr);
            vkDestroyPipeline(base, pipeline, nullptr);
            vkDestroyRenderPass(base, renderPass, nullptr);
        }
    }
};

static Core::UniquePointer<VulkanDummy> dummy;

bool Vulkan::init() {
    dummy = new VulkanDummy();
    return dummy->init();
}

void Vulkan::render() {
    if(dummy->render()) {
        dummy->shouldWait = true;
    }
}

void Vulkan::destroy(void) {
    dummy->destroy();
    dummy = nullptr;
}
